package net.hostsharing.hsadminng.hs.office.coopassets;

import net.hostsharing.hsadminng.config.JsonObjectMapperConfiguration;
import net.hostsharing.hsadminng.context.Context;
import net.hostsharing.hsadminng.hs.office.membership.HsOfficeMembershipEntity;
import net.hostsharing.hsadminng.hs.office.membership.HsOfficeMembershipRepository;
import net.hostsharing.hsadminng.hs.office.partner.HsOfficePartnerEntity;
import net.hostsharing.hsadminng.mapper.StrictMapper;
import net.hostsharing.hsadminng.persistence.EntityManagerWrapper;
import net.hostsharing.hsadminng.rbac.test.JsonBuilder;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.WebMvcTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.context.annotation.Import;
import org.springframework.http.MediaType;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;

import java.util.Optional;
import java.util.UUID;
import java.util.function.Function;

import static net.hostsharing.hsadminng.rbac.test.JsonBuilder.jsonObject;
import static net.hostsharing.hsadminng.rbac.test.JsonMatcher.lenientlyEquals;
import static org.hamcrest.Matchers.is;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

@WebMvcTest(HsOfficeCoopAssetsTransactionController.class)
@Import({ StrictMapper.class, JsonObjectMapperConfiguration.class })
@RunWith(SpringRunner.class)
class HsOfficeCoopAssetsTransactionControllerRestTest {

    private static final UUID AVAILABLE_MEMBERSHIP_UUID = UUID.randomUUID();
    public static final HsOfficeMembershipEntity AVAILABLE_MEMBER_ENTITY = HsOfficeMembershipEntity.builder()
            .uuid(AVAILABLE_MEMBERSHIP_UUID)
            .partner(HsOfficePartnerEntity.builder()
                    .partnerNumber(12345)
                    .build())
            .memberNumberSuffix("00")
            .build();
    private static final UUID UNAVAILABLE_MEMBERSHIP_UUID = UUID.randomUUID();
    private static final String UNAVAILABLE_MEMBER_NUMBER = "M-1234699";
    private static final String AVAILABLE_MEMBER_NUMBER = "M-1234600";

    @Autowired
    MockMvc mockMvc;

    @MockBean
    Context contextMock;

    @Autowired
    @SuppressWarnings("unused") // not used in test, but in controller class
    StrictMapper mapper;

    @MockBean
    @SuppressWarnings("unused") // not used in test, but in base-class of StrictMapper
    EntityManagerWrapper em;

    @MockBean
    HsOfficeCoopAssetsTransactionRepository coopAssetsTransactionRepo;

    @MockBean
    HsOfficeMembershipRepository membershipRepo;

    static final String VALID_INSERT_REQUEST_BODY = """
            {
               "membership.uuid": "%s",
               "transactionType": "DEPOSIT",
               "assetValue": 128.00,
               "valueDate": "2022-10-13",
               "reference": "valid reference",
               "comment": "valid comment",
               "adoptingMembership.uuid": null,
               "adoptingMembership.memberNumber": null
            }
            """.formatted(UUID.randomUUID());

    enum BadRequestTestCases {
        MEMBERSHIP_UUID_MISSING(
                requestBody -> requestBody.without("membership.uuid"),
                "[membershipUuid must not be null but is \"null\"]"), // TODO.impl: should be membership.uuid, Spring validation-problem?

        TRANSACTION_TYPE_MISSING(
                requestBody -> requestBody.without("transactionType"),
                "[transactionType must not be null but is \"null\"]"),

        VALUE_DATE_MISSING(
                requestBody -> requestBody.without("valueDate"),
                "[valueDate must not be null but is \"null\"]"),

        ASSETS_VALUE_FOR_DEPOSIT_MUST_BE_POSITIVE(
                requestBody -> requestBody
                        .with("transactionType", "DEPOSIT")
                        .with("assetValue", -64.00),
                "[for DEPOSIT, assetValue must be positive but is \"-64.00\"]"),

        ASSETS_VALUE_FOR_DISBURSAL_MUST_BE_NEGATIVE(
                requestBody -> requestBody
                        .with("transactionType", "DISBURSAL")
                        .with("assetValue", 64.00),
                "[for DISBURSAL, assetValue must be negative but is \"64.00\"]"),

        //TODO: other transaction types

        ADOPTING_MEMBERSHIP_NUMBER_FOR_TRANSFER_MUST_BE_GIVEN_AND_AVAILABLE(
                requestBody -> requestBody
                        .with("transactionType", "TRANSFER")
                        .with("assetValue", -64.00)
                        .with("adoptingMembership.memberNumber", UNAVAILABLE_MEMBER_NUMBER),
                "adoptingMembership.memberNumber='M-1234699' not found or not accessible"),

        ADOPTING_MEMBERSHIP_UUID_FOR_TRANSFER_MUST_BE_GIVEN_AND_AVAILABLE(
                requestBody -> requestBody
                        .with("transactionType", "TRANSFER")
                        .with("assetValue", -64.00)
                        .with("adoptingMembership.uuid", UNAVAILABLE_MEMBERSHIP_UUID.toString()),
                "adoptingMembership.uuid='" + UNAVAILABLE_MEMBERSHIP_UUID + "' not found or not accessible"),

        ASSETS_VALUE_MUST_NOT_BE_NULL(
                requestBody -> requestBody
                        .with("transactionType", "REVERSAL")
                        .with("assetValue", 0.00),
                "[assetValue must not be 0 but is \"0.00\"]"),

        REFERENCE_MISSING(
                requestBody -> requestBody.without("reference"),
                "[reference must not be null but is \"null\"]"),

        REFERENCE_TOO_SHORT(
                requestBody -> requestBody.with("reference", "12345"),
                "[reference size must be between 6 and 48 but is \"12345\"]"),

        REFERENCE_TOO_LONG(
                requestBody -> requestBody.with("reference", "0123456789012345678901234567890123456789012345678"),
                "[reference size must be between 6 and 48 but is \"0123456789012345678901234567890123456789012345678\"]");

        private final Function<JsonBuilder, JsonBuilder> givenBodyTransformation;
        private final String expectedErrorMessage;

        BadRequestTestCases(
                final Function<JsonBuilder, JsonBuilder> givenBodyTransformation,
                final String expectedErrorMessage) {
            this.givenBodyTransformation = givenBodyTransformation;
            this.expectedErrorMessage = expectedErrorMessage;
        }

        String givenRequestBody() {
            return givenBodyTransformation.apply(jsonObject(VALID_INSERT_REQUEST_BODY)).toString();
        }
    }

    @ParameterizedTest
    @EnumSource(BadRequestTestCases.class)
    void respondWithBadRequest(final BadRequestTestCases testCase) throws Exception {
//        assumeThat(testCase == ADOPTING_MEMBERSHIP_NUMBER_FOR_TRANSFER_MUST_BE_GIVEN_AND_AVAILABLE).isTrue();

        // when
        mockMvc.perform(MockMvcRequestBuilders
                        .post("/api/hs/office/coopassetstransactions")
                        .header("current-subject", "superuser-alex@hostsharing.net")
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(testCase.givenRequestBody())
                        .accept(MediaType.APPLICATION_JSON))

                // then
                .andExpect(jsonPath("message", is("ERROR: [400] " + testCase.expectedErrorMessage)))
                .andExpect(jsonPath("statusCode", is(400)))
                .andExpect(jsonPath("statusPhrase", is("Bad Request")))
                .andExpect(status().is4xxClientError());
    }

    enum SuccessfullyCreatedTestCases {

        ADOPTING_MEMBERSHIP_NUMBER_FOR_TRANSFER_MUST_BE_GIVEN_AND_AVAILABLE(
                requestBody -> requestBody
                        .with("transactionType", "TRANSFER")
                        .with("assetValue", -64.00)
                        .with("adoptingMembership.memberNumber", AVAILABLE_MEMBER_NUMBER),
                """
                {
                    "transactionType": "TRANSFER",
                    "assetValue": -64.00,
                    "adoptionAssetTx": {
                        "transactionType": "ADOPTION",
                        "assetValue": 64.00
                    },
                    "reversalAssetTx": null,
                    "transferAssetTx": null,
                    "revertedAssetTx": null
                }
                """),

        ADOPTING_MEMBERSHIP_UUID_FOR_TRANSFER_MUST_BE_GIVEN_AND_AVAILABLE(
                requestBody -> requestBody
                        .with("transactionType", "TRANSFER")
                        .with("assetValue", -64.00)
                        .with("adoptingMembership.uuid", AVAILABLE_MEMBERSHIP_UUID.toString()),
                """
                {
                    "transactionType": "TRANSFER",
                    "assetValue": -64.00,
                    "adoptionAssetTx": {
                        "transactionType": "ADOPTION",
                        "assetValue": 64.00
                    },
                    "transferAssetTx": null,
                    "revertedAssetTx": null,
                    "reversalAssetTx": null
                }
                """);

        private final Function<JsonBuilder, JsonBuilder> givenBodyTransformation;
        private final String expectedResponseBody;

        SuccessfullyCreatedTestCases(
                final Function<JsonBuilder, JsonBuilder> givenBodyTransformation,
                final String expectedResponseBody) {
            this.givenBodyTransformation = givenBodyTransformation;
            this.expectedResponseBody = expectedResponseBody;
        }

        String givenRequestBody() {
            return givenBodyTransformation.apply(jsonObject(VALID_INSERT_REQUEST_BODY)).toString();
        }
    }

    @ParameterizedTest
    @EnumSource(SuccessfullyCreatedTestCases.class)
    void respondWithSuccessfullyCreated(final SuccessfullyCreatedTestCases testCase) throws Exception {

        // when
        mockMvc.perform(MockMvcRequestBuilders
                        .post("/api/hs/office/coopassetstransactions")
                        .header("current-subject", "superuser-alex@hostsharing.net")
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(testCase.givenRequestBody())
                        .accept(MediaType.APPLICATION_JSON))

                // then
                .andExpect(jsonPath("$", lenientlyEquals(testCase.expectedResponseBody)))
                .andExpect(status().is2xxSuccessful());
    }

    @BeforeEach
    void initMocks() {
        final var availableMemberNumber = Integer.valueOf(AVAILABLE_MEMBER_NUMBER.substring("M-".length()));
        when(membershipRepo.findMembershipByMemberNumber(availableMemberNumber)).thenReturn(AVAILABLE_MEMBER_ENTITY);
        when(membershipRepo.findByUuid(AVAILABLE_MEMBERSHIP_UUID)).thenReturn(Optional.of(AVAILABLE_MEMBER_ENTITY));
        when(coopAssetsTransactionRepo.save(any(HsOfficeCoopAssetsTransactionEntity.class)))
                .thenAnswer(invocation -> {
                            final var entity = (HsOfficeCoopAssetsTransactionEntity) invocation.getArgument(0);
                            if (entity.getUuid() == null) {
                                entity.setUuid(UUID.randomUUID());
                            }
                            return entity;
                        }
                );
    }
}
